# -*- coding: utf-8 -*-

# TODO: After switching to Python 3, this can be rewritten using the
# Python bindings for FL instead of subprocess/piping.

import re
from lxml import etree as ET
import subprocess
import util

SENTTAG = "s"
WORDTAG = "w"

# Sufficiently random material that FreeLing will always treat as one
# token. It signals to the sender that begin/end of processed data has
# been reached in the output.
START = "27345327645267453684527684"
END   = "27345327645267453684527685"


def fl_proc(in_file, out_file, conf_file, lang, sent_end="Fp", slevel=""):
    """ Read an XML or text document and process the text with FreeLing.
    Works for English, French, German, Italian, Portuguese, Russian, Spanish.
    - conf_file is a language specific FreeLing CFG file
    - lang is the two-letter language code of the language to be analyzed
    - sent_end is the tag that marks the end of a sentence
    - slevel should be set if the material already has sentence-level annotations
    """
    # Init the external FreeLingTool.
    fl = subprocess.Popen(  ['analyze', '-f', conf_file],
                            stdout=subprocess.PIPE,
                            stdin=subprocess.PIPE)
    # Parse document into tree.
    try:
        tree = ET.parse(in_file)
        root = tree.getroot()

    # Convert text document to XML.
    except ET.XMLSyntaxError:
        f = open(in_file, 'r').read()
        text = '<text>\n' + f + '\n</text>'
        root = ET.fromstring(text)
        tree = ET.ElementTree(root)

    # Iterate through tree and FreeLingize all text nodes.
    if slevel:
        for node in root.iterfind(".//" + slevel):
            fl_exec(fl, node, lang, sent_end, slevel)
    else:
        for node in root.iter():
            if node.text:
                fl_exec(fl, node, lang, sent_end, slevel)

    # Kill running subprocess
    util.system.kill_process(fl)

    # Write file.
    tree.write(out_file, encoding="UTF-8")


def fl_exec(fl, node, lang, sent_end, slevel):
    """Function that sends one chunk of material to FreeLing and gets the
    analysis back, using pipes. Does sentence segmentation unless slevel is set."""
    if slevel:
        # Collect stuff from under this node.
        # First, the text from this node.
        if node.text is not None:
            rawtext = node.text.encode("utf8")
            # And remove the old text.
            node.text = ""
        else:
            rawtext = ""

        # Text from the childs.
        # Also remove the childs.
        for child in node.iter():
            if child.text is not None:
                rawtext += " " + child.text.encode("utf8")
        node.clear()

    else:
        rawtext = node.text.encode("utf8")
        node.text = ""

        # We assume that at this stage, all sentences are within XML tags.
        # Hence, we start a new <s> for each existing block of text sent to FL.
        newsnode = ET.Element(SENTTAG)

    # Send material with begin/end markers to FreeLing via pipe.
    fl.stdin.write(START + '\n' + rawtext + '\n' + END + '\n')
    fl.stdin.close()

    # Forward to the BEGIN marker.
    readin = ""
    while not re.match(START, readin):
        readin = fl.stdout.readline()

    # Read everything until the END marker occurs.
    while True:
        readin = fl.stdout.readline()
        if re.match(END, readin):
            if slevel:
                break
            else:
                # Close the current sentence and attach to original
                # annotation node (usually <link>).
                if len(list(newsnode)) > 0:
                    node.append(newsnode)
                break

        if len(readin) > 0:
            # ... and process it. We only need the first three fields.
            fields = readin.split(" ")
            if len(fields) >= 3:

                # Add stuff as XML nodes
                # Create new node.
                msd = fields[2].decode("utf8")
                pos = util.msd_to_pos.convert(msd, lang)

                newnode = ET.Element(WORDTAG, {"pos": pos, "msd": msd, "lemma": fields[1].decode("utf8")})
                newnode.text = fields[0].decode("utf8")

                # Append new node.
                if slevel:
                    node.append(newnode)
                else:
                    newsnode.append(newnode)

                if not slevel:
                    # If this ends a sentence, close <s>, attach, and create new <s>.
                    if fields[2].decode("utf8") == sent_end:
                        if len(list(newsnode)) > 0:
                            node.append(newsnode)
                        newsnode = ET.Element(SENTTAG)


if __name__ == "__main__":
    util.run.main(fl_proc)
